import numpy as np
import json
import os
from NuRadioReco.utilities import units, io_utilities
import gzip
from radiotools import helper as hp
from radiotools import coordinatesystems as cs
from scipy import constants
import logging
import pickle
logger = logging.getLogger('NuRadioReco.antennapattern')

# config = ConfigParser.RawConfigParser()
# config_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'antennamodels.cfg')
# if not os.path.exists(config_path):
#     logger.error("antennamodels.cfg does not exists. You need to create this file manually from the antennamodels.cfg.sample file and add the path to the antenna models svn reop")
#     raise OSError
# config.read(config_path)
path_to_antennamodels = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'AntennaModels')


def interpolate_linear(x, x0, x1, y0, y1, interpolation_method='complex'):
    """
    helper function to linearly interpolate between two complex numbers

    Parameters:
    ------
    x: float
        the requested position
    x0, y0: float, complex float
        the first data point
    x1, y1: float, complex float
        the second data point
    interpolation_method: string
        specifies if interpolation is in
        * complex (default) i.e. real and imaginary part
        * magnitude and phase

    Returns: compex float
        the interpolated value
    """
    if (x0 == x1):
        return y0
    if(interpolation_method == 'complex'):
        return y0 + (y1 - y0) * (x - x0) / (x1 - x0)
    elif(interpolation_method == 'magphase'):  # interpolate magnitude and phase
        mag0 = np.abs(y0)
        mag1 = np.abs(y1)
        phase0 = np.angle(y0)
        phase1 = np.angle(y1)
        phase0, phase1 = np.unwrap([phase0, phase1])
        mag = mag0 + (mag1 - mag0) * (x - x0) / (x1 - x0)
        phase = phase0 + (phase1 - phase0) * (x - x0) / (x1 - x0)
        y = mag * np.exp(1j * phase)
        return y
    else:
        logger.error("interpolation mode {} not implemented".format(interpolation_method))
        raise NotImplementedError


def interpolate_linear_vectorized(x, x0, x1, y0, y1, interpolation_method='complex'):
    """
    Same as `interpolate_linear` but all parameters can be vectors

    """
    x = np.array(x)
    mask = x0 != x1
    result = np.zeros_like(x, dtype=np.complex)
    denominator = x1 - x0
    if(interpolation_method == 'complex'):
        result[mask] = y0[mask] + (y1[mask] - y0[mask]) * (x[mask] - x0[mask]) / denominator[mask]
    elif(interpolation_method == 'magphase'):  # interpolate magnitude and phase
        mag0 = np.abs(y0[mask])
        mag1 = np.abs(y1[mask])
        phase0 = np.angle(y0[mask])
        phase1 = np.angle(y1[mask])
        phase0, phase1 = np.unwrap([phase0, phase1])
        mag = mag0 + (mag1 - mag0) * (x[mask] - x0[mask]) / denominator[mask]
        phase = phase0 + (phase1 - phase0) * (x[mask] - x0[mask]) / denominator[mask]
        result[mask] = mag * np.exp(1j * phase)
    else:
        logger.error("interpolation mode {} not implemented".format(interpolation_method))
        raise NotImplementedError
    result[~mask] = y0[~mask]
    return result


def get_group_delay(vector_effective_length, df):
    """
    helper function to calculate the group delay from the vector effecitve length

    Parameters:
    ----------
    vector_effective_length: complex float
        the vector effective length of an antenna
    df: float
        the size of a frequency bin

    Returns: float (the group delay)


    """
    return -np.diff(np.unwrap(np.angle(vector_effective_length))) / df / units.ns / 2 / np.pi


def parse_WIPLD_file(ad1, ra1, orientation, gen_num=1, s_paramateres=[1, 1]):
    """
    reads in WIPLD data

    Parameters
    ----------
    ad1: string
        path to ad1 file
    ra1: string
        path to radiation pattern file
    orientation: string
        path to orientation file
    gen_num: int
        which antenna (one or two) to pull from
    s_parameters: list of 2 ints
        determines which s-parametr to extract (ex: [1,2] extracts S_12 parameter).

    Returns: all parameters of the files
    """
    boresight, tines = np.loadtxt(orientation, delimiter=',')
    orientation_theta, orientation_phi = hp.cartesian_to_spherical(*boresight)
    rotation_theta, rotation_phi = hp.cartesian_to_spherical(*tines)

    ad1_data = np.loadtxt(ad1, comments='>')
    S_1 = ad1_data[:, 1]
    S_2 = ad1_data[:, 2]
    mask = (S_1 == s_paramateres[0]) & (S_2 == s_paramateres[1])
    ff = ad1_data[:, 0][mask] * units.GHz
    Re_Z = ad1_data[:, 5][mask] * units.ohm
    Im_Z = ad1_data[:, 6][mask] * units.ohm
    Z = Re_Z + 1j * Im_Z

    Re_S = ad1_data[:, 7][mask]
    Im_S = ad1_data[:, 8][mask]
    S = Re_S + 1j * Im_S
    with open(ra1, 'r') as fin:
        ff2 = []
        phis = []
        thetas = []
        Ephis = []
        Ethetas = []
        gains = []
        f = None
        skip = False
        for line in fin.readlines():
            if(line.strip().startswith('>')):
                skip = False
                if int(line.split()[3]) != gen_num:
                    skip = True
                else:
                    logger.debug(line.split())
                f = float(line.split()[4])
            else:
                if skip:
                    continue
                ff2.append(f * units.GHz)
                phi, theta, ReEphi, ImEphi, ReEtheta, ImEtheta, gain, gaindb = line.split()
                phis.append(float(phi))
                thetas.append(float(theta))
                Ephis.append(float(ReEphi) + 1j * float(ImEphi))
                Ethetas.append(float(ReEtheta) + 1j * float(ImEtheta))
                gains.append(float(gain))

        if not np.array_equal(ff, np.unique(np.array(ff2))):
            logger.error("error in parsing WIPLD simulation, frequencies of ad1 and ra1 files do not match!")
            return None
        logger.debug(np.unique(np.array(phis)))
        logger.debug(np.unique(np.array(thetas)))
        return orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff, Z, S, np.array(ff2), np.deg2rad(np.array(phis)), np.deg2rad(np.array(thetas)), np.array(Ephis), np.array(Ethetas), np.array(gains)


def preprocess_WIPLD_old(path, gen_num=1, s_paramateres=[1, 1]):
    """
    preprocesses WIPLD file

    this function implements the older insufficient calculation of the vector effective length. This VEL only
    relates the incident electric field to the open circuit voltage and not the voltage in a 50 Ohm system.

    Parameters
    ----------
    path: string
        path to folder containing ad1, ra1, and orientation files.
    gen_num: int
        which antenna (one or two) to pull from
    s_parameters: list of 2 ints
        determines which s-parametr to extract (ex: [1,2] extracts S_12 parameter).

    Returns:
        * orientation theta: boresight direction (zenith angle, 0deg is the zenith, 180deg is straight down)
        * orientation phi: boresight direction (azimuth angle counting from East counterclockwise)
        * rotation theta: rotation of the antenna, is perpendicular to 'orientation', for LPDAs: vector in plane of tines pointing away from connector
        * rotation phi: rotation of the antenna, is perpendicular to 'orientation', for LPDAs: vector in plane of tines pointing away from connector
        * ff2: array of frequencies
        * theta: zenith angle of inicdent electric field
        * phi: azimuth angle of incident electric field
        * H_phi: the complex vector effective length of the ePhi polarization component
        * H_theta: the complex vector effective length of the eTheta polarization component
    """
    from scipy.interpolate import interp1d
    c = constants.c * units.m / units.s
    Z_0 = 119.9169 * np.pi * units.ohm
    split = os.path.split(os.path.dirname(path))
    name = split[1]
    path = split[0]

    orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff, Z, S, ff2, phi, theta, Iphi, Itheta, gains = parse_WIPLD_file(os.path.join(path, name, '{}.ad1'.format(name)),
                                                                                                                   os.path.join(path, name, '{}.ra1'.format(name)),
                                                                                                                   os.path.join(path, name, '{}.orientation'.format(name)),
                                                                                                                   gen_num=gen_num, s_paramateres=s_paramateres)

    theta = 0.5 * np.pi - theta  # 90deg - theta because in WIPL D the theta angle is defined differently

    # sort with increasing frequency, increasing phi, and increasing theta
    index = np.lexsort((theta, phi, ff2))
    ff2 = ff2[index]
    phi = phi[index]
    theta = theta[index]
    Iphi = Iphi[index]
    Itheta = Itheta[index]

    get_Z = interp1d(ff, Z, kind='nearest')
    wavelength = c / ff2
    H_phi = (2 * wavelength * get_Z(ff2) * Iphi) / (Z_0) / 1j
    H_theta = (2 * wavelength * get_Z(ff2) * Itheta) / (Z_0) / 1j

    return orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff2, theta, phi, H_phi, H_theta

#     H = wavelength * (np.real(get_Z(ff2)) / (np.pi * Z_0)) ** 0.5 * gains ** 0.5


def save_preprocessed_WIPLD_old(path):
    """
    saves preprocessed WIPLD files to a pickle file

    Parameters
    ----------
    path: string
        path to folder containing ad1, ra1, and orientation files.
    """
    orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff2, theta, phi, H_phi, H_theta = preprocess_WIPLD_old(path)
    split = os.path.split(os.path.dirname(path))
    name = split[1]
    path = split[0]
    output_filename = '{}.pkl'.format(os.path.join(path, name, name))
    with open(output_filename, 'wb') as fout:
        logger.info('saving output to {}'.format(output_filename))
        pickle.dump([orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff2, theta, phi, H_phi, H_theta], fout, protocol=4)


def preprocess_WIPLD(path, gen_num=1, s_paramateres=[1, 1]):
    """
    preprocesses WIPLD file

    this function implements the older insufficient calculation of the vector effective length. This VEL only
    relates the incident electric field to the open circuit voltage and not the voltage in a 50 Ohm system.

    Parameters
    ----------
    path: string
        path to folder containing ad1, ra1, and orientation files.
    gen_num: int
        which antenna (one or two) to pull from
    s_parameters: list of 2 ints
        determines which s-parametr to extract (ex: [1,2] extracts S_12 parameter).

    Returns:
        * orientation theta: boresight direction (zenith angle, 0deg is the zenith, 180deg is straight down)
        * orientation phi: boresight direction (azimuth angle counting from East counterclockwise)
        * rotation theta: rotation of the antenna, is perpendicular to 'orientation', for LPDAs: vector in plane of tines pointing away from connector
        * rotation phi: rotation of the antenna, is perpendicular to 'orientation', for LPDAs: vector in plane of tines pointing away from connector
        * ff2: array of frequencies
        * theta: zenith angle of inicdent electric field
        * phi: azimuth angle of incident electric field
        * H_phi: the complex vector effective length of the ePhi polarization component
        * H_theta: the complex vector effective length of the eTheta polarization component
    """
    from scipy.interpolate import interp1d
    c = constants.c * units.m / units.s
    Z_0 = 119.9169 * np.pi * units.ohm
    split = os.path.split(os.path.dirname(path))
    name = split[1]
    path = split[0]

    orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff, Z, S, ff2, phi, theta, Iphi, Itheta, gains = parse_WIPLD_file(os.path.join(path, name, '{}.ad1'.format(name)),
                                                                                                                   os.path.join(path, name, '{}.ra1'.format(name)),
                                                                                                                   os.path.join(path, name, '{}.orientation'.format(name)),
                                                                                                                   gen_num=gen_num, s_paramateres=s_paramateres)

    theta = 0.5 * np.pi - theta  # 90deg - theta because in WIPL D the theta angle is defined differently

    # sort with increasing frequency, increasing phi, and increasing theta
    index = np.lexsort((theta, phi, ff2))
    ff2 = ff2[index]
    phi = phi[index]
    theta = theta[index]
    Iphi = Iphi[index]
    Itheta = Itheta[index]

#     get_Z = interp1d(ff, Z, kind='nearest')
    get_S = interp1d(ff, S, kind='nearest')
    wavelength = c / ff2
    V = 1 * units.V
    Z_L = 50 * units.ohm
    H_phi = wavelength * (1 + get_S(ff2)) * Iphi * Z_L / (Z_0) / 1j / V
    H_theta = wavelength * (1 + get_S(ff2)) * Itheta * Z_L / (Z_0) / 1j / V

#     H = wavelength * (np.real(get_Z(ff2)) / (np.pi * Z_0)) ** 0.5 * gains ** 0.5
    return orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff2, theta, phi, H_phi, H_theta

#     output_filename = '{}.pkl'.format(os.path.join(path, name, name))
#     with open(output_filename, 'wb') as fout:
#         logger.info('saving output to {}'.format(output_filename))
#         pickle.dump([orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff2, theta, phi, H_phi, H_theta], fout, protocol=4)


def save_preprocessed_WIPLD(path):
    """
    saves preprocessed WIPLD files to a pickle file

    Parameters
    ----------
    path: string
        path to folder containing ad1, ra1, and orientation files.
    """
    orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff2, theta, phi, H_phi, H_theta = preprocess_WIPLD(path)
    split = os.path.split(os.path.dirname(path))
    name = split[1]
    path = split[0]
    output_filename = '{}.pkl'.format(os.path.join(path, name, name))
    with open(output_filename, 'wb') as fout:
        logger.info('saving output to {}'.format(output_filename))
        pickle.dump([orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff2, theta, phi, H_phi, H_theta], fout, protocol=4)


def save_preprocessed_WIPLD_forARA(path):
    """
    this function saves the realized gain in an ARASim readable format

    Parameters
    ----------
    path: string
        path to folder containing ad1, ra1, and orientation files.
    """
    from scipy.interpolate import interp1d
    c = constants.c * units.m / units.s
    Z_0 = 119.9169 * np.pi * units.ohm
    split = os.path.split(os.path.dirname(path))
    name = split[1]
    path = split[0]

    orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff, Z, S, ff2, phi, theta, Iphi, Itheta, gains = parse_WIPLD_file(os.path.join(path, name, '{}.ad1'.format(name)),
                                                                                                                   os.path.join(path, name, '{}.ra1'.format(name)),
                                                                                                                   os.path.join(path, name, '{}.orientation'.format(name)))

    theta = 0.5 * np.pi - theta  # 90deg - theta because in WIPL D the theta angle is defined differently

    # sort with increasing frequency, increasing phi, and increasing theta
    index = np.lexsort((theta, phi, ff2))
    ff2 = ff2[index]
    phi = phi[index]
    theta = theta[index]
    Iphi = Iphi[index]
    Itheta = Itheta[index]

    wavelength = c / ff2
    V = 1 * units.V
    Z_L = 50 * units.ohm

    get_S = interp1d(ff, S, kind='nearest')
    Gr = gains * (1 - np.abs(get_S(ff2)) ** 2)
    H_phi = wavelength * (1 + get_S(ff2)) * Iphi * Z_L / (Z_0) / 1j / V
    H_theta = wavelength * (1 + get_S(ff2)) * Itheta * Z_L / (Z_0) / 1j / V

    output_filename = '{}.ara'.format(os.path.join(path, name, name))
    with open(output_filename, 'w') as fout:
        for f in sorted(np.unique(ff2)):
            fout.write("freq : {} MHz\n".format(f / units.MHz))
            fout.write("SWR : ???\n")
            fout.write("Theta   Phi      Gain(dB)          Gain          Phase(deg)\n")
            mask = ff2 == f
            for i in range(np.sum(mask)):
                fout.write("{:.4f} {:.4f} {:.4g} {:.4g} {:.2f} {:.2f}\n".format(theta[mask][i] / units.deg,
                                                                       phi[mask][i] / units.deg,
                                                                       0,
                                                                       Gr[mask][i],
                                                                       np.angle(H_theta[mask][i]) / units.deg,
                                                                       np.angle(H_phi[mask][i]) / units.deg))


def get_pickle_antenna_response(path):
    """
    opens and return the pickle file containing the preprocessed WIPL-D antenna simulation
    If the pickle file is not present on the local file system, or if the file is outdated (verified via a sha1 hash sum),
    the file will be downloaded from a central data server


    Parameters:
    ----------
    path: string
        the path to the pickle file

    """

    download_file = False

    # check if gziped pickle file already exists
    if(not os.path.exists(path)):
        logger.warning("antenna pattern {} does not exist, file will be downloaded".format(path))
        download_file = True

    if(os.path.exists(path)):
        BUF_SIZE = 65536 * 2 ** 4  # lets read stuff in 64kb chunks!
        import hashlib
        import json
        sha1 = hashlib.sha1()
        with open(path, 'rb') as f:
            while True:
                data = f.read(BUF_SIZE)
                if not data:
                    break
                sha1.update(data)

        antenna_directory = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(antenna_directory, 'antenna_models_hash.json'), 'r') as fin:
            antenna_hashs = json.load(fin)
            if(os.path.basename(path) in antenna_hashs.keys()):
                if(sha1.hexdigest() != antenna_hashs[os.path.basename(path)]):
                    logger.warning("antenna model {} has changed on the server. downloading newest version...".format(os.path.basename(path)))
                    download_file = True
            else:
                logger.warning("no hash sum of {} available, skipping up-to-date check".format(os.path.basename(path)))

    if(download_file):
        # does not exist yet -> download file
        import requests
        antenna_pattern_name = os.path.splitext(os.path.basename(path))[0]
        URL = 'http://arianna.ps.uci.edu/~arianna/data/AntennaModels/{name}/{name}.pkl'.format(name=antenna_pattern_name)

        folder = os.path.dirname(path)
        if(not os.path.exists(folder)):
            os.makedirs(folder)
        logger.info("downloading antenna pattern {} from {}. This can take a while...".format(antenna_pattern_name, URL))
        r = requests.get(URL)
        if (r.status_code != requests.codes.ok):
            logger.error("error in download of antenna model")
            raise IOError
        with open(path, "wb") as code:
            code.write(r.content)
        logger.warning("...download finished.")

#         # does not exist yet -> precalculating WIPLD simulations from raw WIPLD output
#         preprocess_WIPLD(path)
    res = io_utilities.read_pickle(path, encoding='bytes')
    return res


def parse_AERA_XML_file(path):
    import xml.etree.ElementTree as ET

    if not os.path.exists(path):
        logger.error("AERA antenna file {} not found".format(path))
        raise OSError

    antenna_file = open(path, "rb")

    antenna_data = "<antenna>" + antenna_file.read() + "</antenna>"  # add pseudo root element

    # get root element
    root = ET.fromstring(antenna_data)

    # get frequencies and angles
    frequencies_node = root.find("./frequency")
    frequencies = np.array(frequencies_node.text.strip().split(), dtype=np.float) * units.MHz

    theta_node = root.find("./theta")
    thetas = np.array(theta_node.text.strip().split(), dtype=np.float) * units.deg

    phi_node = root.find("./phi")
    phis = np.array(phi_node.text.strip().split(), dtype=np.float) * units.deg

    n_freqs = len(frequencies)
    n_angles = len(phis)

    # get amplitude and phase
    theta_amps = np.zeros((n_freqs, n_angles))
    theta_phases = np.zeros((n_freqs, n_angles))
    phi_amps = np.zeros((n_freqs, n_angles))
    phi_phases = np.zeros((n_freqs, n_angles))

    for iFreq, freq in enumerate(frequencies / units.MHz):
        freq_string = "%.2f" % freq

        theta_amp_node = root.find("./EAHTheta_amp[@idfreq='%s']" % freq_string)

        # check string
        if(theta_amp_node is None):
            freq_string = "%.1f" % freq

        theta_amp_node = root.find("./EAHTheta_amp[@idfreq='%s']" % freq_string)
        theta_amps[iFreq] = np.array(theta_amp_node.text.strip().split(), dtype=np.float) * units.m

        theta_phase_node = root.find("./EAHTheta_phase[@idfreq='%s']" % freq_string)
        theta_phases[iFreq] = np.deg2rad(np.array(theta_phase_node.text.strip().split(" "), dtype=np.float))

        phi_amp_node = root.find("./EAHPhi_amp[@idfreq='%s']" % freq_string)
        phi_amps[iFreq] = np.array(phi_amp_node.text.strip().split(), dtype=np.float) * units.m

        phi_phase_node = root.find("./EAHPhi_phase[@idfreq='%s']" % freq_string)
        phi_phases[iFreq] = np.deg2rad(np.array(phi_phase_node.text.strip().split(), dtype=np.float))

    return frequencies, phis, thetas, phi_amps, phi_phases, theta_amps, theta_phases


def preprocess_AERA(path):

    frequencies, phis, thetas, phi_amps, phi_phases, theta_amps, theta_phases = parse_AERA_XML_file(path)

    n_freqs = len(frequencies)
    n_angles = len(phis)

    def P2R(magnitude, phase):
        return magnitude * np.exp(1j * phase)

    VEL_thetas = P2R(theta_amps, theta_phases)
    VEL_phis = P2R(phi_amps, phi_phases)

    # (angle) -> (freq * angle)
    thetas = np.tile(thetas, n_freqs)
    phis = np.tile(phis, n_freqs)

    # (freq) -> (freq * angles)
    ff = np.repeat(frequencies, n_angles)

    # sort with increasing frequency, increasing phi, and increasing theta
    index = np.lexsort((thetas, phis, ff))
    VEL_thetas = VEL_thetas.flatten()[index]
    VEL_phis = VEL_phis.flatten()[index]

    # (angle) -> (freq * angle)
    theta = np.tile(thetas, n_freqs)[index]
    phi = np.tile(phis, n_freqs)[index]

    # to avoid issues when deviding throw H (H=0 is ignored)
    # |H| < 0.1 should not happen between 30 - 80 MHz
    H_phi = np.where(np.abs(VEL_phis) > 0.01, VEL_phis, 0)
    H_theta = np.where(np.abs(VEL_thetas) > 0.01, VEL_thetas, 0)

    # values for a upwards pointing LPDA with the arm aligned to the magnetic field
    orientation_theta, orientation_phi, rotation_theta, rotation_phi = 0 * units.deg, 0 * units.deg, 90 * units.deg, 90 * units.deg

    fname = os.path.split(os.path.basename(path))[1].replace('.xml', '')
    output_filename = '{}_InfAir.pkl'.format(os.path.join(path_to_antennamodels, fname, fname))

    directory = os.path.dirname(output_filename)
    if not os.path.exists(directory):
        os.makedirs(directory)

    with open(output_filename, 'wb') as fout:
        logger.info('saving output to {}'.format(output_filename))
        pickle.dump([orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff, theta, phi, H_phi, H_theta], fout, protocol=4)


def parse_ARA_file(ara):
    """
    Helper function that parses the ARAsim ASCII files containig antenna responses

    Parameters:
    ----------
    ara: string
        path to the file

    Returns:
        * ff: array of floats
            frequencies
        * thetas: array of floats
            zenith angle of inicdent electric field
        * phis: array of floats
            azimuth angle of inicdent electric field
        * gains: array of floats
            corresponding linear gain values
        * phases: array of floats
            corresponding phases
     """
    with open(ara, 'r') as fin:
        ff = []
        phis = []
        thetas = []
        gains = []
        phases = []
        f = None
        tmp_phi0_lines = []
        for line in fin.readlines():
            if(line.strip().startswith('freq')):
                # add phi = 360deg = 0deg to data structure (to allow for a automated interpolation
                f = float(line.replace(" ", "").replace("freq", "").replace(":", "").replace("MHz", ""))
                continue
            if(line.strip().startswith('SWR')):
                continue
            if(line.strip().startswith('Theta')):
                continue
            ff.append(f * units.MHz)
            theta, phi, gaindB, gain, phase = line.split()
            if(float(phi) == 0):
                tmp_phi0_lines.append(line)
            phis.append(float(phi) * units.deg)
            thetas.append(float(theta) * units.deg)
            gains.append(float(gain))
            phases.append(float(phase) * units.deg)
            if(float(phi) == 355 and float(theta) == 180):
                for i, tline in enumerate(tmp_phi0_lines):
                    ff.append(f * units.MHz)
                    theta, phi, gaindB, gain, phase = tline.split()
                    if(i == 0):
                        logger.debug("{} {} {} {} {} {}".format(f, theta, phi, gaindB, gain, phase))
                    phis.append(360. * units.deg)
                    thetas.append(float(theta) * units.deg)
                    gains.append(float(gain))
                    phases.append(float(phase) * units.deg)
                tmp_phi0_lines = []

        return np.array(ff), np.array(phis), np.array(thetas), np.array(gains), np.array(phases)


def preprocess_ARA(path):
    """
    preprocess an antenna pattern in the ARASim ASCII file format. The vector effective length is calculated and
    the output is saved to the NuRadioReco pickle format.

    Parameters:
    ----------
    path: string
        the path to the file

    """
    c = constants.c * units.m / units.s
    Z_0 = 119.9169 * np.pi
    split = os.path.split(os.path.dirname(path))
    name = split[1]
    path = split[0]
    orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff, phi, theta, gain, phase = parse_ARA_file(os.path.join(path, name, '{}.txt'.format(name)),
                                                                                                 os.path.join(path, name, '{}.orientation'.format(name)))

    wavelength = c / ff
    H_theta = wavelength * (50 / (np.pi * Z_0)) ** 0.5 * gain ** 0.5  # * np.exp(1j * phase)  ## do not use phases, this will screw up the interpolation
    H_phi = H_theta * 1e-3
    output_filename = '{}.pkl'.format(os.path.join(path, name, name))
    with open(output_filename, 'wb') as fout:
        logger.info('saving output to {}'.format(output_filename))
        pickle.dump([orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff, theta, phi, H_phi, H_theta], fout, protocol=4)


def parse_HFSS_file(hfss):
    """
    Helper function that parses the HFSS files containig antenna responses

    Parameters:
    ----------
    hfss: string
        path to the file

    Returns:
        * ff: array of floats
            frequencies
        * thetas: array of floats
            zenith angle of inicdent electric field
        * phis: array of floats
            azimuth angle of inicdent electric field
        * magnitudes_theta: array of floats
            corresponding logarithmic magnitude values theta component
        * magnitudes_phi: array of floats
            corresponding logarithmic magnitude values phi component
        * phases_phi: array of floats
            corresponding phases phi component
        * phases_theta: array of floats
            corresponding phases theta component
     """
    ff, phi, theta, mag_phi, mag_theta, phase_phi, phase_theta = [], [], [], [], [], [], []
    import re

    with open(hfss, 'r') as csv_file:

        for j, row in enumerate(csv_file.readlines()):

            if j == 0:
                array_names = row.split(',')
            else:
                array = row.split(',')
                for i in range(len(array_names)):
                    if 'Freq' in array_names[i]:
                        freq = array[i]
                    if 'log10(mag(rEPhi))' in array_names[i]:
                        mag_phi.append(float(array[i]))
                        ff.append(float(freq) * units.MHz)

                        p = re.search("Phi='(.+?)deg'", array_names[i])
                        t = re.search("Theta='(.+?)deg'", array_names[i])
                        phi.append(np.deg2rad(int(p.group(1))))
                        theta.append(np.deg2rad(int(t.group(1))))
                    if 'log10(mag(rETheta))' in array_names[i]:
                        mag_theta.append(float(array[i]))
                    if 'ang_rad(rEPhi)' in array_names[i]:
                        phase_phi.append(float(array[i]))
                    if 'ang_rad(rETheta)' in array_names[i]:
                        phase_theta.append(float(array[i]))

        for i in range(len(np.unique(ff)) + 1):
            for arr in [theta, mag_theta, mag_phi, phase_theta, phase_phi, ff, phi]:
                arr[(i - 1) * len(ff) / len(np.unique(ff)):i * len(ff) / len(np.unique(ff))] = [x for _, x in sorted(zip(phi[(i - 1) * len(ff) / len(np.unique(ff)):i * len(ff) / len(np.unique(ff))], arr[(i - 1) * len(ff) / len(np.unique(ff)):i * len(ff) / len(np.unique(ff))]), key=lambda pair: pair[0])]

        return np.array(ff), np.array(phi), np.array(theta), np.array(mag_phi), np.array(mag_theta), np.array(phase_phi), np.array(phase_theta)


def preprocess_HFSS(path):

    """
    preprocess an antenna pattern in the HFSS file format. The vector effective length is calculated and the output is saved in the NuRadioReco pickle format.

    The vector effective length calculation still needs to be verified.

    The frequencies, theta, phi, magnitude theta, magnitude phi, phase theta and phase phi are read from the csv file and than ordered according to the NuRadioReco format.


    Parameters:
    ----------
    path: string
        the path to the file

    """

    split = os.path.split(os.path.dirname(path))
    name = split[1]
    path = split[0]

    ff, phi, theta, mag_phi, mag_theta, phase_phi, phase_theta = parse_HFSS_file((os.path.join(path, name, '{}.csv'.format(name))))
    mag_theta = 10 ** (mag_theta / 10)
    mag_phi = 10 ** (mag_phi / 10)
    gain_theta = 4.0 * np.pi * (mag_theta ** 2) / (2 * 120 * np.pi)
    gain_phi = 4.0 * np.pi * (mag_phi ** 2) / (2 * 120 * np.pi)
    c = constants.c * units.m / units.s
    Z_0 = 119.9169 * np.pi
    wavelength = c / np.array(ff)
    n_index = 1.78

    H_theta = wavelength / n_index ** 0.5 * (50 / (4 * np.pi * Z_0)) ** 0.5 * gain_theta ** 0.5 * np.exp(1j * phase_theta)
    H_phi = wavelength / n_index ** 0.5 * (50 / (4 * np.pi * Z_0)) ** 0.5 * gain_phi ** 0.5 * np.exp(1j * phase_phi)

    orientation_theta = 0
    orientation_phi = 0
    rotation_theta = 0
    rotation_phi = 0

    output_filename = '{}.pkl'.format(os.path.join(path, name, name))

    with open(output_filename, 'wb') as fout:
        logger.info('saving output to {}'.format(output_filename))
        pickle.dump([orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff, theta, phi, H_phi, H_theta], fout, protocol=4)


def preprocess_XFDTD(path):
    """
    preprocess an antenna pattern in the XFDTD file format. The vector effective length is calculated and
    the output is saved to the NuRadioReco pickle format.

    Parameters:
    ----------
    path: string
        the path to the file

    """
    split = os.path.split(os.path.dirname(path))
    name = split[1]
    path = split[0]

    import yaml
    with open(os.path.join(path, name, '{}.yaml'.format(name))) as fin:
        info = yaml.load(fin)
        orientation_theta, orientation_phi = hp.cartesian_to_spherical(*info['boresight_direction'])
        rotation_theta, rotation_phi = hp.cartesian_to_spherical(*info['orientation'])
        n_index = info['n']

        c = constants.c * units.m / units.s
        Z_0 = 119.9169 * np.pi
        ff, phi, theta, gain, phase = parse_ARA_file(os.path.join(path, name, '{}.txt'.format(name)))
        wavelength = c / ff
        H = wavelength / n_index ** 0.5 * (50 / (4 * np.pi * Z_0)) ** 0.5 * gain ** 0.5 * np.exp(1j * phase)
        if(info['type'] == 'Vpol'):
            H_theta = H
            H_phi = H * 1e-6
        elif(info['type'] == 'Hpol'):
            H_theta = H * 1e-6
            H_phi = H
        else:
            logger.error("antenna type {} not understood".format(info['type']))
            raise NotImplementedError("antenna type {} not understood".format(info['type']))

        output_filename = '{}.pkl'.format(os.path.join(path, name, name))
        with open(output_filename, 'wb') as fout:
            logger.info('saving output to {}'.format(output_filename))
            pickle.dump([orientation_theta, orientation_phi, rotation_theta, rotation_phi, ff, theta, phi, H_phi, H_theta], fout, protocol=4)


class AntennaPatternBase():
    """
    base class of utility class that handles access and buffering to antenna pattern
    """

    def _get_antenna_rotation(self, orientation_theta, orientation_phi, rotation_theta, rotation_phi):
        """

        Parameters:
        ----------

        """
        # define orientation of wiplD antenna simulation (in ARIANNA CS)
        e1 = hp.spherical_to_cartesian(self._orientation_theta, self._orientation_phi)  # boresight direction
        e2 = hp.spherical_to_cartesian(self._rotation_theta, self._rotation_phi)  # vector perpendicular to tine plane
        e3 = np.cross(e1, e2)
        E = np.array([e1, e2, e3])
        if(np.linalg.norm(e3) < 0.9):
            logger.error("orientation of antenna not properly defined in WIPL-D orientation file")
            raise AssertionError("orientation of antenna not properly defined in WIPL-D orientation file")

        # get normal vectors for antenne orientation in field (in ARIANNA CS)
        a1 = hp.spherical_to_cartesian(orientation_theta, orientation_phi)
        a2 = hp.spherical_to_cartesian(rotation_theta, rotation_phi)
        a3 = np.cross(a1, a2)
        A = np.array([a1, a2, a3])
        if(np.linalg.norm(a3) < 0.9):
            logger.error("orientation of antenna not properly defined detector description")
            raise AssertionError("orientation of antenna not properly defined detector description")
        from numpy.linalg import inv

        return np.matmul(inv(E), A)

    def _get_theta_and_phi(self, zenith, azimuth, orientation_theta, orientation_phi, rotation_theta, rotation_phi):
        """
        transform zenith and azimuth angle in ARIANNA coordinate system to the WIPLD coordinate system.
        In addition the orientation of the antenna as deployed in the field is taken into account.

        Parameters:
        ----------
        """

        rot = self._get_antenna_rotation(orientation_theta, orientation_phi, rotation_theta, rotation_phi)

        incoming_direction = hp.spherical_to_cartesian(zenith, azimuth)
        incoming_direction_WIPLD = np.dot(rot, incoming_direction.T).T
        theta, phi = hp.cartesian_to_spherical(*incoming_direction_WIPLD)
        if(zenith == 180 * units.deg):
            logger.debug(incoming_direction)
            logger.debug(rot)
            logger.debug(incoming_direction_WIPLD)
#         theta = 0.5 * np.pi - theta  # in wipl D the elevation is defined with 0deg being in the x-y plane
#         theta = hp.get_normalized_angle(theta)
#         phi = hp.get_normalized_angle(phi)

        logger.debug("zen/az {:.0f} {:.0f} transform to {:.0f} {:.0f}".format(zenith / units.deg,
                                                                              azimuth / units.deg,
                                                                              theta / units.deg,
                                                                              phi / units.deg))
        return theta, phi

    def get_antenna_response_vectorized(self, freq, zenith, azimuth, orientation_theta, orientation_phi, rotation_theta, rotation_phi):
        """
        get the antenna response for a specific frequency, zenith and azimuth angle

        All angles are specified in the ARIANNA coordinate system. All units are in ARIANNA default units

        Parameters
        ----------
        freq : float or array of floats
            frequency
        zenith : float
            zenith angle of incoming signal direction
        azimuth : float
            azimuth angle of incoming signal direction
        orientation_theta: float 
            boresight direction (zenith angle, 0deg is the zenith, 180deg is straight down)
        orientation_phi: float 
            boresight direction (azimuth angle counting from East counterclockwise)
        rotation_theta: float
            rotation of the antenna, is perpendicular to 'orientation', for LPDAs: vector in plane of tines pointing away from connector
        rotation_phi: float 
            rotation of the antenna, is perpendicular to 'orientation', for LPDAs: vector in plane of tines pointing away from connector

        Returns
        -------
        VEL: dictonary of complex arrays
            theta and phi component of the vector effective length, both components
            are complex floats or arrays of complex floats
            of the same length as the frequency input
        """
        if self._notfound:
            VEL = {}
            VEL['theta'] = np.ones(len(freq), dtype=np.complex)
            VEL['phi'] = np.ones(len(freq), dtype=np.complex)
            return VEL

        if(isinstance(freq, (float, int))):
            freq = np.array([freq])
        theta, phi = self._get_theta_and_phi(zenith, azimuth, orientation_theta, orientation_phi, rotation_theta, rotation_phi)

        Vtheta_raw, Vphi_raw = self._get_antenna_response_vectorized_raw(freq, theta, phi)

        # now rotate the raw theta and phi component of the VEL into the ARIANNA coordinate system.
        # As the theta and phi angles are differently defined in WIPLD and ARIANNA, also the orientation of the
        # eTheta and ePhi unit vectors are different.
        cstrans = cs.cstrafo(zenith=theta, azimuth=phi)
        V_xyz_raw = cstrans.transform_from_onsky_to_ground(np.array([np.zeros(Vtheta_raw.shape[0]), Vtheta_raw, Vphi_raw]))
        rot = self._get_antenna_rotation(orientation_theta, orientation_phi, rotation_theta, rotation_phi)
        from numpy.linalg import inv
        V_xyz = np.dot(inv(rot), V_xyz_raw)

        cstrans2 = cs.cstrafo(zenith=zenith, azimuth=azimuth)
        V_onsky = cstrans2.transform_from_ground_to_onsky(V_xyz)
        VEL = {}
        VEL['theta'] = V_onsky[1]
        VEL['phi'] = V_onsky[2]
        return VEL


class AntennaPattern(AntennaPatternBase):
    """
    utility class that handles access and buffering to simulated antenna pattern
    """

    def __init__(self, antenna_model, path=path_to_antennamodels,
                 interpolation_method='complex'):
        """

        Parameters
        ----------
        antenna_model: string
            name of antenna model
        path: string
            path to folder containing the antenna models
        interpolation_mode: string
            specify in which domain the interpolation should be performed, can be either
            * 'complex' (default) interpolate real and imaginary part of vector effective length
            * 'magphase' interpolate magnitude and phase of vector effective length
        """

        self._name = antenna_model
        self._interpolation_method = interpolation_method
        from time import time
        t = time()
        filename = os.path.join(path, antenna_model, "{}.pkl".format(antenna_model))
        self._notfound = False
        try:
            self._orientation_theta, self._orientation_phi, self._rotation_theta, self._rotation_phi, \
                    ff, thetas, phis, H_phi, H_theta = get_pickle_antenna_response(filename)

        except IOError:
            self._notfound = True
            logger.warning("antenna response for {} not found".format(antenna_model))
            return

        self.frequencies = np.unique(ff)
        self.frequency_lower_bound = self.frequencies[0]
        self.frequency_upper_bound = self.frequencies[-1]

        self.theta_angles = np.unique(thetas)
        self.theta_lower_bound = self.theta_angles[0]
        self.theta_upper_bound = self.theta_angles[-1]
        logger.debug("{} thetas from {} to {}".format(len(self.theta_angles), self.theta_lower_bound, self.theta_upper_bound))

        self.phi_angles = np.unique(phis)
        self.phi_lower_bound = self.phi_angles[0]
        self.phi_upper_bound = self.phi_angles[-1]
        logger.debug("{} phis from {} to {}".format(len(self.phi_angles), self.phi_lower_bound, self.phi_upper_bound))

        self.n_freqs = len(self.frequencies)
        self.n_theta = len(self.theta_angles)
        self.n_phi = len(self.phi_angles)

        self.VEL_phi = H_phi
        self.VEL_theta = H_theta

        # additional consistency check
        for iFreq, freq in enumerate(self.frequencies):
            for iPhi, phi in enumerate(self.phi_angles):
                for iTheta, theta in enumerate(self.theta_angles):
                    index = self._get_index(iFreq, iTheta, iPhi)

                    if (phi != phis[index]):
                        logger.error("phi angle has changed during theta loop {0}, {1}".format(
                                                phi / units.deg, phis[index] / units.deg))
                        raise Exception("phi angle has changed during theta loop")

                    if (theta != thetas[index]):
                        logger.error("theta angle has changed during theta loop {0}, {1}".format(
                                                theta / units.deg, thetas[index] / units.deg))
                        raise Exception("theta angle has changed during theta loop")

                    if (freq != ff[index]):
                        logger.error("frequency has changed {0}, {1}".format(
                                                        freq, ff[index]))
                        raise Exception("frequency has changed")

        logger.warning('loading antenna file {} took {:.0f} seconds'.format(antenna_model, time() - t))

    def _get_index(self, iFreq, iTheta, iPhi):
        """
        """
        return iFreq * self.n_theta * self.n_phi + iPhi * self.n_theta + iTheta

    def _get_antenna_response_vectorized_raw(self, freq, theta, phi):
        """
        get vector effective length in WIPLD coordinate system
        """
        while phi < self.phi_lower_bound:
            phi += 2 * np.pi
        while phi > self.phi_upper_bound:
            phi -= 2 * np.pi

        if(hp.is_equal(theta, self.theta_upper_bound, rel_precision=1e-5)):
            theta = self.theta_upper_bound
        if(hp.is_equal(theta, self.theta_lower_bound, rel_precision=1e-5)):
            theta = self.theta_lower_bound
        if(((phi < self.phi_lower_bound) or (phi > self.phi_upper_bound)) or
           ((theta < self.theta_lower_bound) or (theta > self.theta_upper_bound))):
            logger.debug(self._name)
            logger.debug("theta bounds {0} ,{1}, {2}".format(self.theta_lower_bound, theta, self.theta_upper_bound))
            logger.debug("phi bounds {0} ,{1}, {2}".format(self.phi_lower_bound, phi, self.phi_upper_bound))
            logger.warning("theta, phi or frequency out of range, returning (0,0j)")
            logger.debug("{0},{1},{2}".format(freq, self.frequency_lower_bound, self.frequency_upper_bound))
            return (0, 0)

        if(self.theta_upper_bound == self.theta_lower_bound):
            iTheta_lower = 0
            iTheta_upper = 0
        else:
            iTheta_lower = np.array(np.floor((theta - self.theta_lower_bound) / (self.theta_upper_bound - self.theta_lower_bound) * (self.n_theta - 1)), dtype=np.int)
            iTheta_upper = np.array(np.ceil((theta - self.theta_lower_bound) / (self.theta_upper_bound - self.theta_lower_bound) * (self.n_theta - 1)), dtype=np.int)
        theta_lower = self.theta_angles[iTheta_lower]
        theta_upper = self.theta_angles[iTheta_upper]
        if(self.phi_upper_bound == self.phi_lower_bound):
            iPhi_lower = 0
            iPhi_upper = 0
        else:
            iPhi_lower = np.array(np.floor((phi - self.phi_lower_bound) / (self.phi_upper_bound - self.phi_lower_bound) * (self.n_phi - 1)), dtype=np.int)
            iPhi_upper = np.array(np.ceil((phi - self.phi_lower_bound) / (self.phi_upper_bound - self.phi_lower_bound) * (self.n_phi - 1)), dtype=np.int)
        phi_lower = self.phi_angles[iPhi_lower]
        phi_upper = self.phi_angles[iPhi_upper]

        iFrequency_lower = np.array(np.floor((freq - self.frequency_lower_bound) / (self.frequency_upper_bound - self.frequency_lower_bound) * (self.n_freqs - 1)), dtype=np.int)
        iFrequency_upper = np.array(np.ceil((freq - self.frequency_lower_bound) / (self.frequency_upper_bound - self.frequency_lower_bound) * (self.n_freqs - 1)), dtype=np.int)
        # handling frequency out of bound cases properly
        out_of_bound_freqs_low = freq < self.frequency_lower_bound
        out_of_bound_freqs_high = freq > self.frequency_upper_bound
        iFrequency_lower[out_of_bound_freqs_low] = 0  # set all out of bound frequencies to its minimum/maximum possible value
        iFrequency_lower[out_of_bound_freqs_high] = 0  # set all out of bound frequencies to its minimum/maximum possible value
        iFrequency_upper[out_of_bound_freqs_low] = self.n_freqs - 1
        iFrequency_upper[out_of_bound_freqs_high] = self.n_freqs - 1
        frequency_lower = self.frequencies[iFrequency_lower]
        frequency_upper = self.frequencies[iFrequency_upper]

        # lower frequency bound
        # theta low
        VELt_freq_low_theta_low = interpolate_linear(
            phi, phi_lower, phi_upper,
            self.VEL_theta[self._get_index(iFrequency_lower, iTheta_lower, iPhi_lower)],
            self.VEL_theta[self._get_index(iFrequency_lower, iTheta_lower, iPhi_upper)],
            self._interpolation_method)
        VELp_freq_low_theta_low = interpolate_linear(
            phi, phi_lower, phi_upper,
            self.VEL_phi[self._get_index(iFrequency_lower, iTheta_lower, iPhi_lower)],
            self.VEL_phi[self._get_index(iFrequency_lower, iTheta_lower, iPhi_upper)],
            self._interpolation_method)

        # theta up
        VELt_freq_low_theta_up = interpolate_linear(
            phi, phi_lower, phi_upper,
            self.VEL_theta[self._get_index(iFrequency_lower, iTheta_upper, iPhi_lower)],
            self.VEL_theta[self._get_index(iFrequency_lower, iTheta_upper, iPhi_upper)],
            self._interpolation_method)
        VELp_freq_low_theta_up = interpolate_linear(
            phi, phi_lower, phi_upper,
            self.VEL_phi[self._get_index(iFrequency_lower, iTheta_upper, iPhi_lower)],
            self.VEL_phi[self._get_index(iFrequency_lower, iTheta_upper, iPhi_upper)],
            self._interpolation_method)

        VELt_freq_low = interpolate_linear(theta, theta_lower,
                                           theta_upper,
                                           VELt_freq_low_theta_low,
                                           VELt_freq_low_theta_up,
                                           self._interpolation_method)
        VELp_freq_low = interpolate_linear(theta, theta_lower,
                                           theta_upper,
                                           VELp_freq_low_theta_low,
                                           VELp_freq_low_theta_up,
                                           self._interpolation_method)

        # upper frequency bound
        # theta low
        VELt_freq_up_theta_low = interpolate_linear(
            phi, phi_lower, phi_upper,
            self.VEL_theta[self._get_index(iFrequency_upper, iTheta_lower, iPhi_lower)],
            self.VEL_theta[self._get_index(iFrequency_upper, iTheta_lower, iPhi_upper)],
            self._interpolation_method)
        VELp_freq_up_theta_low = interpolate_linear(
            phi, phi_lower, phi_upper,
            self.VEL_phi[self._get_index(iFrequency_upper, iTheta_lower, iPhi_lower)],
            self.VEL_phi[self._get_index(iFrequency_upper, iTheta_lower, iPhi_upper)],
            self._interpolation_method)

        # theta up
        VELt_freq_up_theta_up = interpolate_linear(
            phi, phi_lower, phi_upper,
            self.VEL_theta[self._get_index(iFrequency_upper, iTheta_upper, iPhi_lower)],
            self.VEL_theta[self._get_index(iFrequency_upper, iTheta_upper, iPhi_upper)],
            self._interpolation_method)
        VELp_freq_up_theta_up = interpolate_linear(
            phi, phi_lower, phi_upper,
            self.VEL_phi[self._get_index(iFrequency_upper, iTheta_upper, iPhi_lower)],
            self.VEL_phi[self._get_index(iFrequency_upper, iTheta_upper, iPhi_upper)],
            self._interpolation_method)

        VELt_freq_up = interpolate_linear(theta, theta_lower, theta_upper,
                                          VELt_freq_up_theta_low,
                                          VELt_freq_up_theta_up,
                                          self._interpolation_method)
        VELp_freq_up = interpolate_linear(theta, theta_lower, theta_upper,
                                          VELp_freq_up_theta_low,
                                          VELp_freq_up_theta_up,
                                          self._interpolation_method)

        interpolated_VELt = interpolate_linear_vectorized(freq, frequency_lower,
                                                          frequency_upper,
                                                          VELt_freq_low,
                                                          VELt_freq_up,
                                                          self._interpolation_method)
        interpolated_VELp = interpolate_linear_vectorized(freq, frequency_lower,
                                                          frequency_upper,
                                                          VELp_freq_low,
                                                          VELp_freq_up,
                                                          self._interpolation_method)

        # set all out of bound frequencies to zero
        interpolated_VELt[out_of_bound_freqs_low] = 0 + 0 * 1j
        interpolated_VELt[out_of_bound_freqs_high] = 0 + 0 * 1j
        interpolated_VELp[out_of_bound_freqs_low] = 0 + 0 * 1j
        interpolated_VELp[out_of_bound_freqs_high] = 0 + 0 * 1j
        return interpolated_VELt, interpolated_VELp


class AntennaPatternAnalytic(AntennaPatternBase):
    """
    utility class that handles access and buffering to analytic antenna pattern
    """

    def __init__(self, antenna_model, cutoff_freq=50 * units.MHz):
        """

        """
        self._notfound = False
        self._model = antenna_model
        self._cutoff_freq = cutoff_freq
        if(self._model == 'analytic_LPDA'):
            # LPDA dummy model points towards z direction and has its tines in the y-z plane
            logger.info("setting boresight direction")
            self._orientation_theta = 0 * units.deg
            self._orientation_phi = 0 * units.deg
            self._rotation_theta = 90 * units.deg
            self._rotation_phi = 0 * units.deg

    def parametric_phase(self, freq, type='theoretical'):
        """

        """
        if type == 'frontlobe_lpda':
            a = 100 * (freq - 400 * units.MHz) ** 2 - 20
            a[np.where(freq > 400 * units.MHz)] -= 0.00007 * (freq[np.where(freq > 400 * units.MHz)] - 400 * units.MHz) ** 2
        elif type == 'side_lpda':
            a = 40 * (freq - 950 * units.MHz) ** 2 - 40
        elif type == 'back_lpda':
            a = 50 * (freq - 950 * units.MHz) ** 2 - 50
        elif type == "theoretical":
            # ratio of two elements
            tau = 0.75
            # maximum frequency
            f = 1000. * units.MHz
            a = np.pi / np.log(tau) * np.log(freq / f) - 60

        return a

    def _get_antenna_response_vectorized_raw(self, freq, theta, phi, group_delay='frontlobe_lpda'):
        """

        """
        if(self._model == 'analytic_LPDA'):
            """
            Dummy LPDA model.
            Flat gain as function of frequency, no group delay.
            Can be used instead of __get_antenna_response_vectorized_raw
            """
            max_gain_co = 4
            max_gain_cross = 2  # Check whether these values are actually reasonable

            index = np.argmax(freq > self._cutoff_freq)
            Gain = np.ones_like(freq)
            from scipy.signal import hann
            filter = hann(2 * index)
            Gain[:index] = filter[:index]

            # at WIPL-D (1,0,0) Gain max for e_theta (?? I hope)
            # Standard units, deliver H_eff in meters
            Z_0 = constants.physical_constants['characteristic impedance of vacuum'][0] * units.ohm
            Z_ant = 50 * units.ohm

            # Assuming simple cosine, sine falls-off for dummy module
            H_eff_t = np.zeros_like(Gain)
            fmask = freq >= 0
            H_eff_t[fmask] = Gain[fmask] * max_gain_cross * 1 / freq[fmask]
            H_eff_t *= np.cos(theta) * np.sin(phi)
            H_eff_t *= constants.c * units.m / units.s * Z_ant / Z_0 / np.pi

            H_eff_p = np.zeros_like(Gain)
            H_eff_p[fmask] = Gain[fmask] * max_gain_co * 1 / freq[fmask]
            H_eff_p *= np.cos(phi)
            H_eff_p *= constants.c * units.m / units.s * Z_ant / Z_0 / np.pi

            if group_delay != None:
                # add here antenna model with analytic description of typical group delay
                phase = self.parametric_phase(freq, group_delay)

                H_eff_p = H_eff_p.astype(complex)
                H_eff_t = H_eff_t.astype(complex)

                H_eff_p *= np.exp(1j * phase)
                H_eff_t *= np.exp(1j * phase)

            return H_eff_p, H_eff_t


class AntennaPatternProvider(object):
    __instance = None

    def __new__(cls):
        if AntennaPatternProvider.__instance is None:
            AntennaPatternProvider.__instance = object.__new__(cls)
        return AntennaPatternProvider.__instance

    def __init__(self):
        """
        Provider class for antenna pattern. The usage of antenna pattern through this class ensures
        that an antenna pattern is loaded only once into memory which takes a significant time and occupies a
        significant amount of memory.
        """
        self._open_antenna_patterns = {}
        self._antenna_model_replacements = {}

        antenna_directory = os.path.dirname(os.path.abspath(__file__))
        filename = os.path.join(antenna_directory, 'antenna_model_replacements.json')
        if(os.path.exists(filename)):
            with open(filename, 'r') as fin:
                self._antenna_model_replacements = json.load(fin)

    def load_antenna_pattern(self, name, **kwargs):
        """
        loads an antenna pattern and returns the antenna pattern class

        Paramters
        ----------
        name: string
            the name of the antenna pattern
        **kwargs: dict
            key word arguments that are passed to the init function of the `AntennaPattern` class (see
            documentation of this class for further information)
        """
        if(name in self._antenna_model_replacements.keys()):
            if(self._antenna_model_replacements[name] not in self._open_antenna_patterns.keys()):
                logger.warning("local replacement of antenna model requsted: replacing {} with {}".format(name, self._antenna_model_replacements[name]))
            name = self._antenna_model_replacements[name]
        if (name not in self._open_antenna_patterns.keys()):
            if(name.startswith("analytic")):
                self._open_antenna_patterns[name] = AntennaPatternAnalytic(name, **kwargs)
                logger.info("loading analytic antenna model {}".format(name))
            else:
                self._open_antenna_patterns[name] = AntennaPattern(name, **kwargs)
        return self._open_antenna_patterns[name]



from __future__ import absolute_import, division, print_function
import numpy as np
import logging
from NuRadioReco.utilities import fft
try:
    import cPickle as pickle
except ImportError:
    import pickle
logger = logging.getLogger("BaseTrace")


class BaseTrace:

    def __init__(self):
        self._sampling_rate = None
        self._time_trace = None
        self._frequency_spectrum = None
        self.__time_domain_up_to_date = True
        self._trace_start_time = 0

    def get_trace(self):
        """
        returns the time trace. If the frequency spectrum was modified before,
        an ifft is performed automatically to have the time domain representation
        up to date.

        Returns: 1 or N dimensional np.array of floats
            the time trace
        """
        if(not self.__time_domain_up_to_date):
#             logger.debug("time domain is not up to date, calculating FFT on the fly")
            self._time_trace = fft.freq2time(self._frequency_spectrum, self._sampling_rate)
            self.__time_domain_up_to_date = True
            self._frequency_spectrum = None
        return self._time_trace

    def get_frequency_spectrum(self):
        if(self.__time_domain_up_to_date):
#             logger.debug("frequency domain is not up to date, calculating FFT on the fly")
#             logger.debug("time trace has shape {}".format(self._time_trace.shape))
            self._frequency_spectrum = fft.time2freq(self._time_trace, self._sampling_rate)
            self._time_trace = None
#             logger.debug("frequency spectrum has shape {}".format(self._frequency_spectrum.shape))
            self.__time_domain_up_to_date = False
        return self._frequency_spectrum

    def set_trace(self, trace, sampling_rate):
        """
        sets the time trace

        Parameters
        -----------
        trace: np.array of floats
            the time series
        sampling_rate: float
            the sampling rage of the trace, i.e., the inverse of the bin width
        """
        if trace is not None:
            if trace.shape[trace.ndim - 1]%2 != 0:
                raise ValueError('Attempted to set trace with an uneven number ({}) of samples. Only traces with an even number of samples are allowed.'.format(trace.shape[trace.ndim - 1]))
        self.__time_domain_up_to_date = True
        self._time_trace = trace
        self._sampling_rate = sampling_rate
        self._frequency_spectrum = None

    def set_frequency_spectrum(self, frequency_spectrum, sampling_rate):
        self.__time_domain_up_to_date = False
        self._frequency_spectrum = frequency_spectrum
        self._sampling_rate = sampling_rate
        self._time_trace = None

    def get_sampling_rate(self):
        """
        returns the sampling rate of the trace

        Return: float
            sampling rate, i.e., the inverse of the bin width
        """
        return self._sampling_rate

    def get_times(self):
        try:
            length = self.get_number_of_samples()
            times = np.arange(0, length / self._sampling_rate - 0.1/self._sampling_rate, 1. / self._sampling_rate) + self._trace_start_time
            if(len(times) != length):
                logger.error("time array does not have the same length as the trace. n_samples = {:d}, sampling rate = {:.5g}".format(length, self._sampling_rate))
                raise ValueError("time array does not have the same length as the trace")
        except:
            times = np.array([])
        return times

    def set_trace_start_time(self, start_time):
        self._trace_start_time = start_time

    def add_trace_start_time(self, start_time):
        self._trace_start_time += start_time

    def get_trace_start_time(self):
        return self._trace_start_time

    def get_frequencies(self):
        length = self.get_number_of_samples()
        return np.fft.rfftfreq(length, d=(1. / self._sampling_rate))

    def get_hilbert_envelope(self):
        from scipy import signal
        h = signal.hilbert(self.get_trace())
        return np.array([np.abs(h[0]), np.abs(h[1]), np.abs(h[2])])

    def get_hilbert_envelope_mag(self):
        return np.linalg.norm(self.get_hilbert_envelope(), axis=0)

    def get_number_of_samples(self):
        """
        returns the number of samples in the time domain

        Return: int
            number of samples in time domain
        """
        length = 0
        if(self.__time_domain_up_to_date):
            length = self._time_trace.shape[-1]  # returns the correct length independent of the dimension of the array (channels are 1dim, efields are 3dim)
        else:
            length = (self._frequency_spectrum.shape[-1] - 1) * 2
        return length

    def serialize(self):
        data = {'sampling_rate': self.get_sampling_rate(),
                'time_trace': self.get_trace(),
                'trace_start_time': self.get_trace_start_time()}
        return pickle.dumps(data, protocol=4)

    def deserialize(self, data_pkl):
        data = pickle.loads(data_pkl)
        self.set_trace(data['time_trace'], data['sampling_rate'])
        if('trace_start_time' in data.keys()):
            self.set_trace_start_time(data['trace_start_time'])



import numpy as np
from NuRadioReco.modules.base.module import register_run
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import electricFieldParameters as efp
from NuRadioReco.utilities import units
import NuRadioReco.utilities.trace_utilities
import logging
import radiotools.helper as hp
import radiotools.coordinatesystems
import radiotools.atmosphere.models


class cosmicRayEnergyReconstructor:
    """
    Reconstructs the energy of an air shower from its radio signal
    Requires the following modules to be run beforehand:
        - a 10th order Butterworth bandpass filter with passband 80-300 MHz
        - a direction reconstruction
        - the voltageToAnalyticEfieldConverter
    """

    def __init__(self):
        self.logger = logging.getLogger('NuRadioReco.cosmicRayEnergyReconstructor')
        self.__atmosphere = radiotools.atmosphere.models.Atmosphere()
        self.__parametrizations = {
            'mooresbay': {
                'scale': np.array([(442.46, -281.75, 324.96), (394.08, -308.36, 436.30)]),
                'falloff': np.array([(-.1584, -.07943), (.8070, -1.4098)])
            },
            'southpole': {
                'scale': np.array([(976.30, -1213.43, 626.98), (643.39, -667.08, 478.06)]),
                'falloff': np.array([(-.2273, .05627), (1.3372, -2.1653)])
            },
            'auger': {
                'scale': np.array([(229.96, -123.75, 110.51), (214.46, -111.01, 119.18)]),
                'falloff': np.array([(-.1445, -.09820), (.5936, -1.1763)])
            }
        }
        self.__elevations = {  # TODO: This should be changed once we have implemented a proper coordinate system
            'mooresbay': 30.,
            'southpole': 2800.,
            'auger': 1560.
        }

    def begin(self, site='mooresbay'):
        self.__site = site
        if site not in self.__parametrizations.keys():
            self.logger.error('Unsupported site. Please select one of the following: {}'.format(self.__parametrizations.keys()))
            raise ValueError
        self.__parametrization_for_site = self.__parametrizations[site]
        self.__elevation = self.__elevations[site]

    @register_run()
    def run(self, event, station, detector):

        if not station.is_cosmic_ray():
            self.logger.warning('Event is not a cosmic ray!')
        if not station.has_parameter(stnp.zenith) or not station.has_parameter(stnp.azimuth):
            self.logger.error('No incoming direction available. Energy can not be reconstructed!')
            return
        zenith = station.get_parameter(stnp.zenith)
        azimuth = station.get_parameter(stnp.azimuth)
        if zenith < 30.*units.deg:
            self.logger.warning('Zenith angle is smaller than 30deg. Energy reconstruction is likely to be inaccurate!')
        n_efields = len(station.get_electric_fields())
        if n_efields == 0:
            self.logger.error('No E-field found. Please run the voltageToAnalyticEfieldConverter beforehand!')
            return
        if n_efields > 1:
            self.logger.warning('Multiple E-fields were found. Only the first E-field will be used.')
        electric_field = station.get_electric_fields()[0]
        spectrum_slope = electric_field.get_parameter(efp.cr_spectrum_slope)
        alpha = hp.get_angle_to_magnetic_field_vector(zenith, azimuth, self.__site)
        cs = radiotools.coordinatesystems.cstrafo(zenith, azimuth, site=self.__site)
        efield_trace_vxB_vxvxB = cs.transform_to_vxB_vxvxB(cs.transform_from_onsky_to_ground(electric_field.get_trace()))
        efield_trace_vxB_vxvxB[0] /= np.sin(alpha)  # correct energy fluence for effect of angle to magnetic field
        energy_fluence = NuRadioReco.utilities.trace_utilities.get_electric_field_energy_fluence(efield_trace_vxB_vxvxB, electric_field.get_times())
        energy_fluence = np.abs(energy_fluence[0]) + np.abs(energy_fluence[1])
        xmax_distance = self.__atmosphere.get_distance_xmax_geometric(zenith, 750., self.__elevation)  # parametrization is for Xmax of 750g/cm^2
        # find out if we are inside or outside of the Cherenkov ring
        second_order_spectrum_parameter = electric_field.get_parameter(efp.cr_spectrum_quadratic_term)
        if second_order_spectrum_parameter <= spectrum_slope * .1:
            scale_parameter = self.__parametrization_for_site['scale'][0][0] * zenith ** 2 + self.__parametrization_for_site['scale'][0][1] * zenith + self.__parametrization_for_site['scale'][0][0]
            falloff_parameter = self.__parametrization_for_site['falloff'][0][0] * zenith + self.__parametrization_for_site['falloff'][0][1]
        else:
            scale_parameter = self.__parametrization_for_site['scale'][1][0] * zenith ** 2 + self.__parametrization_for_site['scale'][1][1] * zenith + self.__parametrization_for_site['scale'][1][0]
            falloff_parameter = self.__parametrization_for_site['falloff'][1][0] * zenith + self.__parametrization_for_site['falloff'][1][1]
        rec_energy = 1.e18 * np.sqrt(energy_fluence) * (xmax_distance / units.km) / (scale_parameter * np.exp(falloff_parameter * np.abs(spectrum_slope) ** 0.8))
        station.set_parameter(stnp.cr_energy_em, rec_energy)




from NuRadioReco.modules.base.module import register_run
import NuRadioReco.framework.event
import NuRadioReco.framework.station
import NuRadioReco.framework.channel
import ROOT
import numpy as np
from NuRadioReco.utilities import units
from NuRadioReco.framework.parameters import ARIANNAParameters as ARIpar
import datetime
import sys
import time
import os
sys.path.append(os.path.expandvars('$SNS'))
from scripts.online import AriUtils
from scripts.offline import dacs2014
import logging



class readARIANNAData:
    """
    Reads ARIANNA data as preprocessed in snowShovel. Can read RawData, FPNCorrectedData
    and CalibratedData.
    """

    def __init__(self):
        self.logger = logging.getLogger("NuRadioReco.readARIANNAData")

    def begin(self, input_files, trigger_types=None, time_interval=None,
              tree='AmpOutData', run_number=None, event_ids=None,
              random_iterator=False):
        """
        read FPN and gain calibrated ARIANNA data (snowshovel data format)

        Parameters
        ----------
        input_files: array of strings
            a list of input files
        trigger_types: list of strings or None
            only events that have the specified trigger types are read.
            The trigger types can ['thermal', 'forced']. Default is None which
            means that all events are read regardless of their trigger type
        time_interval: [datetime, datetime] or None
            all events outside of the specified time interval are skipped
            Default is None, i.e., all events are read
        tree: string
            low level reconstruction stage. Available options are
             * RawData: the raw ADC counts
             * FPNSubData: FPN subtracted data (still ADC counts but with mean zero)
             * AmpOutData: FPN and gain calibrated data (default)
        run_number: int or None
            run number, all events with a different run number will be skipped.
            Default is None which means that all events are read in
        event_ids: dictionary or None
            specify any combination of run and event ids, all other events will be skipped.
            key is the run id, values are the event ids
            Default is None which means that all events are read in
        random_iterator: bool (default False)
            if True walk through event in a random order
        """
        self.__trigger_types = trigger_types
        self.__time_interval = time_interval
        self.__run_number = run_number
        self.__event_ids = event_ids
        self.data_tree = ROOT.TChain("CalibTree")
        self.config_tree = ROOT.TChain("ConfigTree")

        for input_file in input_files:
            self.logger.info("adding file {}".format(input_file))
            self.data_tree.Add(input_file)
            self.config_tree.Add(input_file)

        self.calwv = ROOT.TSnCalWvData()
        if(tree == 'RawData'):
            self.calwv = ROOT.TSnRawWaveform()
        self.data_tree.SetBranchAddress("{}.".format(tree), ROOT.AddressOf(self.calwv))
#         self.data_tree.SetBranchAddress("AmpOutData.", self.calwv)

        self.readout_config = ROOT.TSnReadoutConfig()

        self.config_tree.SetBranchAddress("ReadoutConfig.", ROOT.AddressOf(self.readout_config))
        self.skipped_events = 0
        self.skipped_events_stop = 0

        self.n_events = self.data_tree.GetEntries()
        self._evt_range = np.arange(self.n_events, dtype=np.int)
        if(random_iterator):
            np.random.shuffle(self._evt_range)
        self.__id_current_event = -1

        n_events_config = self.config_tree.GetEntries()
        self.logger.debug("{} entries in config".format(n_events_config))
        self._config_keys = {}
        for i in range(n_events_config):
            self.config_tree.GetEntry(i)
            stn_id = AriUtils.getStnFromMacAdr(self.config_tree.ConfigMetadata.GetStationId())
            run_num = self.config_tree.ConfigMetadata.GetRunNum()
            seq_num = self.config_tree.ConfigMetadata.GetSeqNum()
            self._config_keys[(stn_id, run_num, seq_num)] = i

        self.__t = time.time()

        return self.n_events

    @register_run()
    def run(self):
        while True:
            self.__id_current_event += 1
            if(self.__id_current_event >= self.n_events):
                # all events processed
                break
            if(self.__id_current_event % 1000 == 0):
                progress = 1. * self.__id_current_event / self.n_events
                eta = 0
                if(self.__id_current_event > 0):
                    eta = (time.time() - self.__t) / self.__id_current_event * (self.n_events - self.__id_current_event) / 60.
                self.logger.warning("reading in event {}/{} ({:.0f}%) ETA: {:.1f} minutes".format(self.__id_current_event, self.n_events,
                                                                         100 * progress, eta))
#             try:
            self.data_tree.GetEntry(self._evt_range[self.__id_current_event])

            evt_time = datetime.datetime.fromtimestamp(self.data_tree.EventHeader.GetUnixTime())
            if(self.__time_interval is not None):
                if(evt_time < self.__time_interval[0]):
                    continue
                if(evt_time > self.__time_interval[1]):
                    continue
            if(self.__trigger_types is not None):
                use_event = False
                for trigger in self.__trigger_types:
                    if(trigger == 'thermal'):
                        if(self.data_tree.EventHeader.IsThermal()):
                            use_event = True
                    if(trigger == 'forced'):
                        if(self.data_tree.EventHeader.IsForced()):
                            use_event = True
                if(use_event is False):
                    self.logger.debug("skipping event because trigger type was not {type}".format(type=self.__trigger_types))
                    continue

            mac = self.data_tree.EventMetadata.GetStationId()
            self._station_id = AriUtils.getStnFromMacAdr(mac)

            evt_number = self.data_tree.EventHeader.GetEvtNum()
            run_number = self.data_tree.EventMetadata.GetRunNum()
            if(self.__run_number is not None):
                if(run_number != self.__run_number):
                    continue
            if(self.__event_ids is not None):
                if(run_number not in self.__event_ids):
                    continue
                if(evt_number not in self.__event_ids[run_number]):
                    continue

            seq_number = self.data_tree.EventMetadata.GetSeqNum()
            try:
                self.config_tree.GetEntry(self._config_keys[(self._station_id, run_number, seq_number)])
            except:
                self.logger.error("no config entry exists for station {}, run {} and sequence {}. Skipping event...".format(self._station_id, run_number, seq_number))
                self.skipped_events += 1
                continue
            # check if config and event sequence are the same
            seq_num_config = self.config_tree.ConfigMetadata.GetSeqNum()
            if(seq_number != seq_num_config):
                raise Exception("seq number in config {} does not match sequence number in event {}".format(seq_num_config, seq_number))

            evt_triggered = self.data_tree.EventHeader.IsThermal()
            evt = NuRadioReco.framework.event.Event(run_number, evt_number)

            if(self.__id_current_event % 1000 == 0):
                self.logger.info("reading in station {station_id} run {run_number} event {evt_number} at time {time}".format(station_id=self._station_id, run_number=run_number, evt_number=evt_number, time=evt_time))
            self.logger.debug("reading in station {station_id} run {run_number} event {evt_number} at time {time}".format(station_id=self._station_id, run_number=run_number, evt_number=evt_number, time=evt_time))

            nChan = ord(self.readout_config.GetNchans())  # convert char to int
            name = self.readout_config.GetTypeName()
            if name == 'Custom':
                self.logger.warning("Event {event} of run {run} is skipped, as ReadoutConfig seems empty".format(event=evt_number, run=run_number))
                self.skipped_events += 1
                continue

            self.sampling_rate = self.readout_config.GetSamplingRate() * units.GHz

            station = NuRadioReco.framework.station.Station(self._station_id)
            station.set_station_time(evt_time)
            station.set_triggered(evt_triggered)
            stop = np.array(self.data_tree.RawData.GetStopSamples())

            # skip events that don't have a proper information of the stop point
            if (stop.size != 0):
                for iCh in range(nChan):
                    channel = NuRadioReco.framework.channel.Channel(iCh)
                    voltage = np.array(self.calwv.GetDataOnCh(iCh)) * units.mV
                    voltage = np.roll(voltage, -stop[0])
                    channel.set_trace(voltage, self.sampling_rate)
                    station.add_channel(channel)
            else:
                self.logger.warning(" Event {event} of run {run} is skipped, no stop point for rolling array!".format(event=evt_number, run=run_number))
                self.skipped_events_stop += 1
                continue

            station.set_ARIANNA_parameter(ARIpar.seq_num, seq_number)
            # read and save start and stop time of a sequence
            start = datetime.datetime.fromtimestamp(self.config_tree.TrigStartClock.GetCurrTime().GetSec())
            stop = datetime.datetime.fromtimestamp(self.config_tree.TrigStopClock.GetCurrTime().GetSec())
            if(start < datetime.datetime(1971, 1, 1)):
                start = None
            if(stop < datetime.datetime(1971, 1, 1)):
                stop = None
            station.set_ARIANNA_parameter(ARIpar.seq_start_time, start)
            station.set_ARIANNA_parameter(ARIpar.seq_stop_time, stop)

            station.set_ARIANNA_parameter(ARIpar.comm_duration, self.config_tree.DAQConfig.GetCommWinDuration() * units.s)
            station.set_ARIANNA_parameter(ARIpar.comm_period, self.config_tree.DAQConfig.GetCommWinPeriod() * units.s)
            station.set_ARIANNA_parameter(ARIpar.l1_supression_value, self.config_tree.DAQConfig.GetL1SingleFreqRatioCut())
            station.set_ARIANNA_parameter(ARIpar.internal_clock_time, self.data_tree.EventHeader.GetDTms() * units.ms)
            dacset = self.config_tree.DAQConfig.GetDacSet().GetDacs()
            dacsV = {}
            for iCh in range(len(dacset)):
                dacsV[iCh] = {}
                for ihl, hl in enumerate(['low', 'high']):
                    dacsV[iCh][hl] = dacs2014.getVth(AriUtils.getBoardFromMacAdr(AriUtils.getMacAdrFromStn(self._station_id)), iCh, dacset[iCh][ihl], hl)
            station.set_ARIANNA_parameter(ARIpar.trigger_thresholds, dacsV)

            evt.set_station(station)
            yield evt
#             except:
#                 self.logger.error("error in reading in event station {station_id} run {run_number} event {evt_number} at time {time}".format(station_id=self._station_id,
#                 run_number=run_number,
#                 evt_number=evt_number,
#                 time=evt_time))
#                 continue

    def end(self):
        if self.skipped_events > 0:
            self.logger.warning("Skipped {} events due to problems in config".format(self.skipped_events))
        if self.skipped_events_stop > 0:
            self.logger.warning("Skipped {} events due to problems in stop bit".format(self.skipped_events_stop))



import dash
import json
import plotly
import numpy as np
from app import app
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from NuRadioReco.utilities import units
from NuRadioReco.eventbrowser.default_layout import default_layout
from NuRadioReco.framework.parameters import channelParameters as chp
import NuRadioReco.eventbrowser.apps.common
import dataprovider
provider = dataprovider.DataProvider()

layout = [
    dcc.Dropdown(id='dropdown-overview-channels',
        options=[],
        multi=True,
        value=[]
    ),
    html.Div(id='channel-overview-properties')
]

@app.callback(Output('dropdown-overview-channels', 'options'),
                [Input('filename', 'value'),
                Input('event-counter-slider', 'value'),
                Input('station-id-dropdown', 'value'),
                Input('station-overview-rec-sim', 'value')],
                [State('user_id', 'children')])
def dropdown_overview_channels(filename, evt_counter, station_id, rec_or_sim, juser_id):
    if filename is None or station_id is None:
        return ''
    user_id = json.loads(juser_id)
    ariio = provider.get_arianna_io(user_id, filename)
    evt = ariio.get_event_i(evt_counter)
    station = evt.get_station(station_id)
    if station is None:
        return []
    options = []
    for channel in station.iter_channels():
        options.append({
            'label': 'Ch. {}'.format(channel.get_id()),
            'value': channel.get_id()
        })
    return options

channel_properties_for_overview = [
    {
        'label': 'Signal to Noise ratio',
        'param': chp.SNR,
        'unit': None
    },{
        'label': 'Max. amplitude [microVolt]',
        'param': chp.maximum_amplitude,
        'unit': units.microvolt
    },{
        'label': 'Max. of Hilbert envelope [microVolt]',
        'param': chp.maximum_amplitude_envelope,
        'unit': units.microvolt
    },{
        'label': 'Cosmic Ray Template Correlations',
        'param': chp.cr_xcorrelations,
        'unit': None
    },{
        'label': 'Neutrino Template Correlations',
        'param': chp.nu_xcorrelations,
        'unit': None
    }
]

@app.callback(Output('channel-overview-properties', 'children'),
                [Input('filename', 'value'),
                Input('event-counter-slider', 'value'),
                Input('station-id-dropdown', 'value'),
                Input('dropdown-overview-channels', 'value')],
                [State('user_id', 'children')])
def channel_overview_properties(filename, evt_counter, station_id, selected_channels, juser_id):
    if filename is None or station_id is None:
        return ''
    user_id = json.loads(juser_id)
    ariio = provider.get_arianna_io(user_id, filename)
    evt = ariio.get_event_i(evt_counter)
    station = evt.get_station(station_id)
    if station is None:
        return []
    reply = []
    for channel_id in selected_channels:
        channel = station.get_channel(channel_id)
        props = NuRadioReco.eventbrowser.apps.common.get_properties_divs(channel, channel_properties_for_overview)
        reply.append(
            html.Div([
            html.Div([
                html.Div('Channel {}'.format(channel.get_id()), className='custom-table-th')
            ],className='custom-table-row'),
            html.Div(props)
        ]))
    return reply



from __future__ import absolute_import, division, print_function  # , unicode_literals
from dash.dependencies import Input, Output, State
import time
import dash
import radiotools.helper as hp
import dash_html_components as html
import dash_core_components as dcc
import plotly.graph_objs as go
import plotly
from plotly import tools
import json
from app import app
import dataprovider
from NuRadioReco.utilities import units
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import channelParameters as chp
from NuRadioReco.framework.parameters import electricFieldParameters as efp
from NuRadioReco.eventbrowser.apps.common import get_point_index
import NuRadioReco.eventbrowser.apps.cosmic_ray_plots.cosmic_ray_skyplot
import NuRadioReco.eventbrowser.apps.cosmic_ray_plots.cosmic_ray_polarization_zenith
import numpy as np
import logging
logger = logging.getLogger('traces')

provider = dataprovider.DataProvider()

layout = html.Div([
    html.Div([
        html.Div([
            html.Div([
                html.Div('Polarization', className='panel-heading'),
                html.Div(NuRadioReco.eventbrowser.apps.cosmic_ray_plots.cosmic_ray_polarization_zenith.layout, className='panel-body')
            ], className='panel panel-default', style={'flex': '1'}),
            html.Div([
                html.Div('Direction Reconstruction', className='panel-heading', style={'display': 'flex'}),
                html.Div(NuRadioReco.eventbrowser.apps.cosmic_ray_plots.cosmic_ray_skyplot.layout, className='panel-body')
            ], className='panel panel-default', style={'flex': '1'})
        ], style={'display': 'flex'})
    ])
])
